#import <driverkit/i386/directDevice.h>
#import <driverkit/i386/IOEISADeviceDescription.h>
#import <driverkit/generalFuncs.h>
#import <driverkit/interruptMsg.h>
#import <kernserv/kern_server_types.h>
#import <kernserv/prototypes.h>
#import "Mididriver.h"

/* We have one "master driver" that manages all cards.  Each card is an
 * instance of Mididriver.  However there is one owner of the "master
 * driver". This has the disadvantage that one app can not use one card 
 * while another app uses another card (unless there's a "Midi Manager").
 * But it has the advantage that time control is global, which
 * means that only a single card need listen to MIDI time code
 * to provide synchronization.
 *
 * Therefore, the "unit" in mididriver_server.c becomes the card.
 * Some cards have multiple ports.  These we call "sub-units".
 */

#if 0
Date: Mon, 16 Nov 92 15:53:14 -0800
From: david (David A. Jaffe)
To: mminnick@next.com
Cc: r@next.com, david
Subject: bug in MIDI driver

Now that I'm actually using the MIDI driver, I've decided that there's
a bug in it.  The problem is that, when in MIDI time code mode,
there's no way to know when the time it returns is actually valid.
For example, I have a tape with some time code recorded.  There are no
full messages-- the time code just starts running.  It's my suspicion
(though I can't prove it without a debug kernel) that when I receive
the "time code started" exception, the driver's idea of the current
time is "half baked".  This is because the time is not knowable until
8 quarter frames have gone by.  The problem is that there's no way for
the user to know when time is valid.  At the moment, I'm just waiting
a bit before asking the time.  But this is an unreliable work-around
because time code could have stopped in the mean time.  (I.e. it could
have given 6 frames and then stopped.)  

We really need a function that returns when time code is valid.  A
conservative addition to the API would be for MIDIGetMTCTime() to
return a new error, "MD_TIME_CODE_UNDEFINED", for this case.

I anticipate it would take me only a couple of hours to implement and test.
#endif

static int _MididriverUnit = 0;

#define MAX_UNITS 8

static Mididriver *driverObjects[MAX_UNITS] = {nil};

@implementation Mididriver

+ (BOOL)probe:deviceDescription
  /*
   * Probe, configure board and init new instance.  This method is 
   * documented in the IODevice spec sheet.
   */
{
    Mididriver *instance;
    IOEISADeviceDescription
      *devDesc = (IOEISADeviceDescription *)deviceDescription;
    if (_MididriverUnit >= MAX_UNITS) {
	IOLog("Mididriver: Too many MIDI devices installed.  Maximum allowed = %d\n",
	      MAX_UNITS);
	return NO;
    }
    instance = [self alloc];
    if (instance == nil) {
      IOLog("Can't allocate Mididriver object.\n");
      return NO;
    }
    /* Check that we have some I/O ports, mapped memory, and 
     * interrupts assigned. */
    if ([devDesc numPortRanges] < 1 || [devDesc numInterrupts] < 1) {
      IOLog("Mididriver: Num port ranges == %d. Num interrupts == %d.\n",
	    [devDesc numPortRanges],[devDesc numInterrupts]);
      [instance free]; 
      return NO;
    }
    /* Perform more device-specific validation, e.g. checking to make 
     * sure the I/O port range is large enough.  Make sure the 
     * hardware is really there. Return NO if anything is wrong. */
    
    return [instance initFromDeviceDescription:devDesc] != nil;
}

- initFromDeviceDescription:deviceDescription
{
  /*
   * Init the new instance.  This method is documented in the i386-specific
   * part of the IODirectDevice spec sheet.
   */
    int err,numRanges;
    int i,baseIO,io1;
    char name[80];
    const IORange *range; 
    /* 
     * If the resources specified in this driver's bundle 
     * (in /usr/Devices/Mididriver.config/*.table) are already reserved,
     * [super initFromDeviceDescription:] will return nil.
     */
//    if ([super initFromDeviceDescription:deviceDescription] == nil)
//    	return nil;
	range = [deviceDescription portRangeList];
	numRanges = [deviceDescription numPortRanges];

    sprintf(name, "Examining: %s%d", "Mididriver", _MididriverUnit);
	baseIO = 0;

	for (i=0;i<numRanges;i++) {
		io1 = range->start;
		IOLog("%s: IO base address=0x%x\n",name,io1);
		if (io1 == 0x300)  {
			baseIO = io1;
			[deviceDescription setPortRangeList:
				(IORange *)range num:(unsigned int)1];
			}
                if (io1 == 0x330)  {
			baseIO = io1;
                        [deviceDescription setPortRangeList:
				(IORange *)range num:(unsigned int)1];
                        }
		range++;
	}	
	if (baseIO == 0)  return nil;
	
	/* sbrandon moved the super initting to here, because I
	 * wanted to try to only reserve the i/o ports that the
	 * MIDI driver is going to use i.e. 0x330 or 0x300. However
	 * although this doesn't fail, nor does it prevent the PnP
	 * system from reserving any extra addresses anyway. Bah.
	 */
	 
	if ([super initFromDeviceDescription:deviceDescription] == nil) {
		IOLog("Failed allocation\n");
       		return nil;
	}
    sprintf(name, "%s%d", "Mididriver", _MididriverUnit);
    IOLog("%s: interrupt=%d, IO base address=0x%x\n",
	  name,
	  [deviceDescription interrupt],
	  baseIO);
    dataPort = baseIO;
    statAndCmdPort = baseIO+1;
    [self setName:name];
    /* Make it possible for MIG interface to find us */
    [self setUnit:_MididriverUnit];
    driverObjects[_MididriverUnit++] = self;
    [self setDeviceKind:"MIDI"]; /* Added Sept. 5, 94 */
    [self setLocation:NULL];
    [self registerDevice];
    if ((err=[self startIOThread]) != IO_R_SUCCESS)
	IOLog("startIOThread failed: %d\n",err);
    return self; 
}

#import "mididriver_server.c"

#define AT_INTERRUPT_LEVEL 0  /* Set to 1 to handle interrupts at true interrupt level */

#if !AT_INTERRUPT_LEVEL
-(void)interruptOccurred
{
    rcvInterrupt([self unit]);
}

#else

static void myIntHandler(void *identity, void *state, 
                         unsigned int arg)
{
  rcvInterrupt(arg);
  return;        
}

- (BOOL) getHandler:(IOEISAInterruptHandler *)handler
       level:(unsigned int *) ipl
       argument:(unsigned int *) arg
       forInterrupt:(unsigned int) localInterrupt
{
    *handler = myIntHandler;
    *ipl = IPLDEVICE;
    *arg = [self unit];
    IOLog("Mididriver: interrupt %d associated with object 0x%x\n",localInterrupt,
	  self);
    return YES;
}
#endif

- (IOReturn)getCharValues   : (unsigned char *)parameterArray
               forParameter : (IOParameterName)parameterName
                      count : (unsigned int *)count
{
  /* 
   * This method is documented in the IODevice spec sheet.
   */
	
  /* Handle any parameters we want to support */
    return [super getCharValues:parameterArray 
	  forParameter:parameterName count:count];
}



@end
